"""
Copyright (C)
This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with this program. If not, see https://tldrlegal.com/license/gnu-general-public-license-v3-(gpl-3)#fulltext. 

For license issues, please contact:

Dr. Bing Ye
Life Sciences Institute
University of Michigan
210 Washtenaw Avenue, Room 5403
Ann Arbor, MI 48109-2216
USA

Email: bingye@umich.edu
"""
from __future__ import annotations

import datetime
import functools
import gc
import operator
import os
from collections import deque
from typing import Sequence, Tuple, Union

import cv2
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sb
from cv2.typing import MatLike
from matplotlib.colorbar import ColorbarBase
from matplotlib.colors import LinearSegmentedColormap, Normalize
from numpy.typing import NDArray
from skimage import exposure
from tensorflow.keras.preprocessing.image import img_to_array


# Constants to store animal lighter vs darker than background
class AnimalVsBg:
    ANIMAL_LIGHTER = 0
    ANIMAL_DARKER = 1
    HARD_TO_TELL = 2


# Use Frame to refer to different representations of a frame
Frame = Union[NDArray[np.uint8], MatLike]


def _extract_background(
    frames: Sequence[MatLike],
    stable_illumination: bool = True,
    animal_vs_bg: int = AnimalVsBg.ANIMAL_LIGHTER,
) -> Frame | None:
    """
    Extracts the background from the given list of frames.

    Args:
        frames: A list of frames to extract the background from.
        stable_illumination: False if background lighting changes for
            optogenetic experiments, else True.
        animal_vs_bg: Whether the animal is lighter or darker than the
            background or whether it's hard to tell; use constants at top of
            tools.py for specific values.

    Returns:
        An NDArray containing the extracted background.

    Raises:
        None
    """

    len_frames = len(frames)

    # Need at least 4 frames to extract background
    if len_frames <= 3:
        return None

    # Convert to ndarray
    frames_arr = np.array(frames, dtype="float32")

    if animal_vs_bg == AnimalVsBg.HARD_TO_TELL:
        if len_frames <= 101:
            return np.uint8(np.median(frames_arr, axis=0))  # type: ignore

        frames_mean = []
        check_frames = []
        mean_overall = frames_arr.mean(0)
        for n in range(0, len_frames - 101, 30):
            frames_temp = frames_arr[n : n + 100]
            mean = frames_temp.mean(0)
            frames_mean.append(frames_temp.mean(0))
            check_frames.append(abs(mean - mean_overall) + frames_temp.std(0))
        frames_mean = np.array(frames_mean, dtype="float32")
        check_frames = np.array(check_frames, dtype="float32")
        background = np.uint8(
            np.take_along_axis(frames_mean, np.argsort(check_frames, axis=0), axis=0)[0]
        )
        del frames_mean
        del check_frames
        del frames_temp
        gc.collect()
        return background  # type: ignore

    if stable_illumination is True:
        return (
            np.uint8(frames_arr.max(0))
            if animal_vs_bg == AnimalVsBg.ANIMAL_DARKER
            else np.uint8(frames_arr.min(0))
        )  # type:ignore

    if len_frames > 101:
        frames_mean = []
        check_frames = []
        for n in range(0, len_frames - 101, 30):
            frames_temp = frames_arr[n : n + 100]
            mean = frames_temp.mean(0)
            frames_mean.append(mean)
            if animal_vs_bg == 1:
                frames_temp_inv = 255 - frames_temp
                check_frames.append(frames_temp_inv.mean(0) + frames_temp_inv.std(0))
            else:
                check_frames.append(mean + frames_temp.std(0))
        frames_mean = np.array(frames_mean, dtype="float32")
        check_frames = np.array(check_frames, dtype="float32")
        background = np.uint8(
            np.take_along_axis(frames_mean, np.argsort(check_frames, axis=0), axis=0)[0]
        )
        del frames_mean
        del check_frames
        del frames_temp
        gc.collect()
        return background  # type: ignore

    return (
        np.uint8(frames_arr.max(0))
        if animal_vs_bg == AnimalVsBg.ANIMAL_DARKER
        else np.uint8(frames_arr.min(0))
    )  # type: ignore


def extract_backgrounds_from_video(
    path_to_video: str,
    delta: float,
    frame_size: tuple[int, int] | None = None,
    stable_illumination: bool = True,
    start_time: float = 0.0,
    end_time: float | None = None,
    animal_vs_bg: int = AnimalVsBg.ANIMAL_LIGHTER,
) -> tuple[Frame, Frame, Frame]:
    """
    Extracts backgrounds from given video.

    Args:
        path_to_video: The path to the video file.
        delta: The factor by which to compare frame brightness. The mean
            brightness of each frame is compared to the mean brightness of the
            initial frame multiplied by and divided by delta to categorize the
            frame as default, low, or high.
        frame_size: The dimensions (width, height) by which to resize the frame
        stable_illumination: False if background lighting changes for
            optogenetic experiments, else True.
        start_time: The beginning of the time window to start extraction.
        end_time: The end of the time window to stop extraction.
        animal_vs_bg: Whether the animal is lighter or darker than the
            background or whether it's hard to tell; use constants at top of
            tools.py for specific values.

    Returns:
        A tuple (default, low, high) where `default` is the default
        background, `low` is the dim-light background, `high` is the
        bright-light background.

    Raises:
        None
    """
    print("Extracting the static background...")

    BG_TYPES = ["default", "low", "high"]
    capture = cv2.VideoCapture(path_to_video)
    fps = capture.get(cv2.CAP_PROP_FPS)
    frame_count = int(capture.get(cv2.CAP_PROP_FRAME_COUNT))
    duration = frame_count / fps
    initial_frame = None
    lower_threshold = None
    upper_threshold = None

    if start_time >= duration:
        print(
            "The beginning time for background extraction is later than the end of the video!"
        )
        print(
            "Will use the 1st second of the video as the beginning time for background extraction!"
        )
        start_time = 0
    if start_time == end_time:
        end_time = start_time + 1

    # Initialize data storage
    frames = {}
    counts = {}
    background_options = {}
    backgrounds = {}
    for bg_type in BG_TYPES:
        frames[bg_type] = deque(maxlen=1000)
        counts[bg_type] = 0
        background_options[bg_type] = deque(maxlen=1000)
        backgrounds[bg_type] = None

    for frame_number in range(1, frame_count + 1):
        _, frame = capture.read()

        reached_start = frame_number >= start_time * fps
        reached_end = (
            end_time is not None and frame_number >= end_time * fps or frame is None
        )

        if reached_end:
            break

        if not reached_start:
            continue

        if frame_size is not None:
            frame = cv2.resize(frame, frame_size, interpolation=cv2.INTER_AREA)

        frame = np.array(frame, dtype=np.uint8)

        if initial_frame is None:
            initial_frame = frame
            lower_threshold = np.mean(initial_frame) / delta
            upper_threshold = np.mean(initial_frame) * delta

        if np.mean(frame) < lower_threshold:  # type: ignore
            frames["low"].append(frame)
            counts["low"] += 1
        elif np.mean(frame) > upper_threshold:  # type: ignore
            frames["high"].append(frame)
            counts["high"] += 1
        else:
            frames["default"].append(frame)
            counts["default"] += 1

        # Extract background if frame buffer has reached 1000 frames
        for bg_type in BG_TYPES:
            if counts[bg_type] == 1001:
                counts[bg_type] = 1
                background = _extract_background(
                    frames[bg_type], stable_illumination, animal_vs_bg
                )
                if background is not None:
                    background_options[bg_type].append(background)

    capture.release()

    for bg_type in BG_TYPES:
        if len(background_options[bg_type]) > 0:
            # Process any remaining frames
            if counts[bg_type] > 600:
                background = _extract_background(
                    frames[bg_type], stable_illumination, animal_vs_bg
                )
                if background is not None:
                    background_options[bg_type].append(background)

            # Pick the best background option
            if len(background_options[bg_type]) == 1:
                backgrounds[bg_type] = background_options[bg_type][0]
            else:
                options = np.array(background_options[bg_type], dtype="uint8")
                if animal_vs_bg == AnimalVsBg.ANIMAL_LIGHTER:
                    backgrounds[bg_type] = options.min(axis=0)
                elif animal_vs_bg == AnimalVsBg.ANIMAL_DARKER:
                    backgrounds[bg_type] = options.max(axis=0)
                elif animal_vs_bg == AnimalVsBg.HARD_TO_TELL:
                    backgrounds[bg_type] = np.median(options, axis=0)
                del options
                gc.collect()
        else:
            backgrounds[bg_type] = _extract_background(
                frames[bg_type], stable_illumination, animal_vs_bg
            )

    if backgrounds["default"] is None:
        background = initial_frame
    if backgrounds["low"] is None:
        backgrounds["low"] = backgrounds["default"]
    if backgrounds["high"] is None:
        backgrounds["high"] = backgrounds["default"]

    print("Background extraction completed!")

    return (backgrounds["default"], backgrounds["low"], backgrounds["high"])  # type: ignore


def load_backgrounds_from_folder(
    folder: str, frame_size: tuple[int, int] | None = None
) -> tuple[Frame, Frame, Frame]:
    """
    Loads background images from given folder, resizing them if required.

    Args:
        folder: The path to the folder containing the background images.
        frame_size: The dimensions (width, height) by which to resize the frame

    Returns:
        A tuple (default, low, high) where `default` is the default
        background, `low` is the dim-light background, `high` is the
        bright-light background.

    Raises:
        None
    """
    default = cv2.imread(os.path.join(folder, "background.jpg"))
    low = cv2.imread(os.path.join(folder, "background_low.jpg"))
    high = cv2.imread(os.path.join(folder, "background_high.jpg"))
    if frame_size is not None:
        default = cv2.resize(default, frame_size, interpolation=cv2.INTER_AREA)
        low = cv2.resize(low, frame_size, interpolation=cv2.INTER_AREA)
        high = cv2.resize(high, frame_size, interpolation=cv2.INTER_AREA)
    return (default, low, high)


def get_stimulation_time(path_to_video: str, delta: float) -> float | None:
    """
    Calculates optogenetic stimulation time from given video.

    Args:
        path_to_video: The path to the video.
        delta: The factor by which to compare frame brightness. The mean
            brightness of each frame is compared to the mean brightness of the
            initial frame multiplied by and divided by delta to categorize the
            frame as default, low, or high.

    Returns:
        The time of the lighting change in seconds, None if no lighting change
        is detected.

    Raises:
        None
    """
    capture = cv2.VideoCapture(path_to_video)
    frame_count = int(capture.get(cv2.CAP_PROP_FRAME_COUNT))
    fps = capture.get(cv2.CAP_PROP_FPS)
    lower_threshold = None
    upper_threshold = None

    for frame_number in range(1, frame_count + 1):
        _, frame = capture.read()

        if frame is None:
            break
        frame = np.array(frame, dtype="uint8")

        if lower_threshold is None or upper_threshold is None:
            lower_threshold = np.mean(frame) / delta
            upper_threshold = np.mean(frame) * delta

        if np.mean(frame) < lower_threshold:  # type: ignore[reportGeneralTypeIssues]
            capture.release()
            return frame_number / fps
        if np.mean(frame) > upper_threshold:  # type: ignore[reportGeneralTypeIssues]
            capture.release()
            return frame_number / fps

    return None


def estimate_animal_area(
    path_to_video: str,
    delta: float,
    background_default: Frame,
    background_low: Frame,
    background_high: Frame,
    animal_number: int,
    frame_size: tuple[int, int] | None = None,
    stim_t: float | None = None,
    start_time: float | None = None,
    duration: float = 10.0,
    animal_vs_bg: int = AnimalVsBg.ANIMAL_LIGHTER,
    kernel_size: int = 3,
) -> float | None:
    """
    Estimates animal size from a video.

    For all frames within the given time window, the background is subtracted
    and morphological transformations (see https://docs.opencv.org/4.7.0/d9/d61/tutorial_py_morphological_ops.html)
    are applied to fill in any holes within the animal mask. The average area
    of all animals in each frame is divided by the number of animals to return
    average animal size in pixels.

    Args:
        path_to_video: The path to the video.
        delta: The factor by which to compare frame brightness. The mean
            brightness of each frame is compared to the mean brightness of the
            initial frame multiplied by and divided by delta to categorize the
            frame as default, low, or high.
        background_default: The default-lighting background.
        background_low: The low-light background.
        background_high: The bright-light background.
        animal_number: The number of animals present in the video.
        frame_size: The dimensions (width, height) by which to resize the frame
        stim_t: The time in seconds of a lighting change in the video, used
            in optogenetic experiments.
        start_time: The time in seconds at which to start animal size
            estimation.
        duration: The duration in seconds of the animal size estimation.
        animal_vs_bg: Whether the animal is lighter or darker than the
            background or whether it's hard to tell; use constants at top of
            tools.py for specific values.
        kernel_size: The kernel size to use for morphological transformations,
            see https://docs.opencv.org/4.7.0/d9/d61/tutorial_py_morphological_ops.html
            for more information

    Returns:
        The area of the animal in pixels after accounting for frame resizing.

    Raises:
        None
    """
    if animal_number == 0:
        print("Animal number is 0. Please enter the correct animal number!")
        return None

    print("Estimating the animal size...")
    print(datetime.datetime.now())

    capture = cv2.VideoCapture(path_to_video)
    fps = capture.get(cv2.CAP_PROP_FPS)
    frame_count = int(capture.get(cv2.CAP_PROP_FRAME_COUNT))

    if start_time is None:
        start_time = stim_t if stim_t is not None else 0.0

    # Cap estimation duration at 30 sec
    if not 0 < duration <= 30:
        duration = 30.0

    end_time = start_time + duration

    min_area = (background_default.shape[1] / 100) * (background_default.shape[0] / 100)
    max_area = (background_default.shape[1] * background_default.shape[0]) * 3 / 4

    background_estimation = background_default
    background_low_estimation = background_low
    background_high_estimation = background_high

    # Make animal lighter than background for morphological transformation
    if animal_vs_bg == AnimalVsBg.ANIMAL_DARKER:
        background_estimation = 255 - background_default  # type: ignore
        background_low_estimation = 255 - background_low  # type: ignore
        background_high_estimation = 255 - background_high  # type: ignore

    lower_threshold = None
    upper_threshold = None
    total_contour_areas = []
    for frame_number in range(1, frame_count + 1):
        _, frame = capture.read()

        reached_start = frame_number >= start_time * fps
        reached_end = (
            end_time is not None and frame_number >= end_time * fps or frame is None
        )
        if reached_end:
            break
        if not reached_start:
            continue

        # Preprocess frame
        if frame_size is not None:
            frame = cv2.resize(frame, frame_size, interpolation=cv2.INTER_AREA)
        frame = np.array(frame, dtype=np.uint8)

        # Make animal brighter than background for morphological transformation
        if animal_vs_bg == AnimalVsBg.ANIMAL_DARKER:
            frame = 255 - frame

        if lower_threshold is None or upper_threshold is None:
            lower_threshold = np.mean(frame) / delta
            upper_threshold = np.mean(frame) * delta

        # Select correct background to use
        if np.mean(frame) < lower_threshold:
            background = background_low_estimation
        elif np.mean(frame) > upper_threshold:
            background = background_high_estimation
        else:
            background = background_estimation

        # Subtract background and convert to grayscale
        if animal_vs_bg == AnimalVsBg.HARD_TO_TELL:
            foreground = cv2.absdiff(frame, background)
        else:
            foreground = cv2.subtract(frame, background)
        foreground = cv2.cvtColor(foreground, cv2.COLOR_BGR2GRAY)

        # Extract contours from frame
        _, thred = cv2.threshold(
            foreground, 0, 255, cv2.THRESH_BINARY + cv2.THRESH_OTSU
        )
        kernel = np.ones((kernel_size, kernel_size), np.uint8)
        thred = cv2.morphologyEx(thred, cv2.MORPH_CLOSE, kernel)
        if animal_vs_bg == AnimalVsBg.HARD_TO_TELL:
            kernel_erode_size = max(kernel_size - 4, 1)
            kernel_erode = np.ones((kernel_erode_size, kernel_erode_size), np.uint8)
            thred = cv2.erode(thred, kernel_erode)
        contours, _ = cv2.findContours(thred, cv2.RETR_LIST, cv2.CHAIN_APPROX_NONE)

        # Get total contour area and append to list
        contour_area = 0
        for contour in contours:
            if min_area < cv2.contourArea(contour) < max_area:
                contour_area += cv2.contourArea(contour)
        total_contour_areas.append(contour_area)

    capture.release()

    print("Estimation completed!")

    if len(total_contour_areas) <= 0:
        print("No animal detected!")
        return None

    animal_area = sum(total_contour_areas) / len(total_contour_areas) / animal_number
    print(f"Single animal size: {animal_area}")
    return animal_area


def crop_frame(frame, contours):
    lfbt = np.array(
        [contours[i].min(0) for i in range(len(contours)) if contours[i] is not None]
    ).min(0)[0]
    x_lf = lfbt[0]
    y_bt = lfbt[1]
    rttp = np.array(
        [contours[i].max(0) for i in range(len(contours)) if contours[i] is not None]
    ).max(0)[0]
    x_rt = rttp[0]
    y_tp = rttp[1]

    w = x_rt - x_lf + 1
    h = y_tp - y_bt + 1

    difference = int(abs(w - h) / 2) + 1

    if w > h:
        y_bt = max(y_bt - difference - 1, 0)
        y_tp = min(y_tp + difference + 1, frame.shape[0])
        x_lf = max(x_lf - 1, 0)
        x_rt = min(x_rt + 1, frame.shape[1])
    if w < h:
        y_bt = max(y_bt - 1, 0)
        y_tp = min(y_tp + 1, frame.shape[0])
        x_lf = max(x_lf - difference - 1, 0)
        x_rt = min(x_rt + difference + 1, frame.shape[1])

    return (y_bt, y_tp, x_lf, x_rt)


def extract_blob(frame, contour, channel=1):
    mask = np.zeros_like(frame)

    cv2.drawContours(mask, [contour], 0, (255, 255, 255), -1)
    mask = cv2.morphologyEx(mask, cv2.MORPH_CLOSE, np.ones((5, 5), np.uint8))

    x, y, w, h = cv2.boundingRect(contour)
    difference = int(abs(w - h) / 2) + 1

    if w > h:
        y_bt = max(y - difference - 1, 0)
        y_tp = min(y + h + difference + 1, frame.shape[0])
        x_lf = max(x - 1, 0)
        x_rt = min(x + w + 1, frame.shape[1])
    else:
        y_bt = max(y - 1, 0)
        y_tp = min(y + h + 1, frame.shape[0])
        x_lf = max(x - difference - 1, 0)
        x_rt = min(x + w + difference + 1, frame.shape[1])

    masked_frame = frame * (mask / 255.0)
    blob = masked_frame[y_bt:y_tp, x_lf:x_rt]
    blob = np.uint8(exposure.rescale_intensity(blob, out_range=(0, 255)))

    if channel == 1:
        blob = cv2.cvtColor(blob, cv2.COLOR_BGR2GRAY)
        blob = img_to_array(blob)

    return blob


def extract_blob_background(
    frame, contours, contour=None, channel=1, background_free=False
):
    (y_bt, y_tp, x_lf, x_rt) = crop_frame(frame, contours)
    if background_free is True:
        mask = np.zeros_like(frame)
        cv2.drawContours(mask, [contour], 0, (255, 255, 255), -1)
        masked_frame = frame * (mask / 255.0)
    else:
        masked_frame = frame
    blob = masked_frame[y_bt:y_tp, x_lf:x_rt]
    blob = np.uint8(exposure.rescale_intensity(blob, out_range=(0, 255)))

    if channel == 1:
        blob = cv2.cvtColor(blob, cv2.COLOR_BGR2GRAY)
        blob = img_to_array(blob)

    return blob


def extract_blob_all(
    frame, y_bt, y_tp, x_lf, x_rt, contours=None, channel=1, background_free=False
):
    if background_free is True:
        mask = np.zeros_like(frame)
        cv2.drawContours(mask, contours, -1, (255, 255, 255), -1)
        masked_frame = frame * (mask / 255.0)
    else:
        masked_frame = frame
    blob = masked_frame[y_bt:y_tp, x_lf:x_rt]
    blob = np.uint8(exposure.rescale_intensity(blob, out_range=(0, 255)))

    if channel == 1:
        blob = cv2.cvtColor(blob, cv2.COLOR_BGR2GRAY)
        blob = img_to_array(blob)

    return blob


def get_inner(masked_frame_gray, contour):
    blur = cv2.GaussianBlur(masked_frame_gray, (3, 3), 0)
    edges = cv2.Canny(blur, 20, 75, apertureSize=3, L2gradient=True)
    cnts, _ = cv2.findContours(edges, cv2.RETR_CCOMP, cv2.CHAIN_APPROX_NONE)

    if len(cnts) > 3:
        inner = sorted(cnts, key=cv2.contourArea, reverse=True)[2:]
    else:
        inner = [contour, contour]

    return inner


def contour_frame(
    frame,
    animal_number,
    background,
    background_low,
    background_high,
    delta,
    contour_area,
    animal_vs_bg=0,
    include_bodyparts=False,
    animation_analyzer=False,
    channel=1,
    kernel=5,
):
    if animal_vs_bg == 1:
        frame = np.uint8(255 - frame)

    if np.mean(frame) < np.mean(background) / delta:
        if animal_vs_bg == 2:
            foreground = cv2.absdiff(frame, background_low)
        else:
            foreground = cv2.subtract(frame, background_low)
    elif np.mean(frame) > delta * np.mean(background):
        if animal_vs_bg == 2:
            foreground = cv2.absdiff(frame, background_high)
        else:
            foreground = cv2.subtract(frame, background_high)
    else:
        if animal_vs_bg == 2:
            foreground = cv2.absdiff(frame, background)
        else:
            foreground = cv2.subtract(frame, background)

    foreground = cv2.cvtColor(foreground, cv2.COLOR_BGR2GRAY)
    thred = cv2.threshold(foreground, 0, 255, cv2.THRESH_BINARY + cv2.THRESH_OTSU)[1]
    thred = cv2.morphologyEx(
        thred, cv2.MORPH_CLOSE, np.ones((kernel, kernel), np.uint8)
    )
    if animal_vs_bg == 2:
        kernel_erode = max(kernel - 4, 1)
        thred = cv2.erode(thred, np.ones((kernel_erode, kernel_erode), np.uint8))
    cnts, _ = cv2.findContours(thred, cv2.RETR_LIST, cv2.CHAIN_APPROX_NONE)

    contours = []
    centers = []
    heights = []
    inners = []
    blobs = []

    if animal_number > 1:
        for i in cnts:
            if contour_area * 0.2 < cv2.contourArea(i) < contour_area * 1.5:
                contours.append(i)
        contours = sorted(contours, key=cv2.contourArea)[-animal_number:]
    else:
        contours = [sorted(cnts, key=cv2.contourArea, reverse=True)[0]]

    if len(contours) > 0:
        for i in contours:
            centers.append(
                (
                    int(cv2.moments(i)["m10"] / cv2.moments(i)["m00"]),
                    int(cv2.moments(i)["m01"] / cv2.moments(i)["m00"]),
                )
            )
            (_, _), (w, h), _ = cv2.minAreaRect(i)
            heights.append(max(w, h))
            if include_bodyparts is True:
                mask = np.zeros_like(frame)
                cv2.drawContours(mask, [i], 0, (255, 255, 255), -1)
                mask = cv2.dilate(mask, np.ones((5, 5), np.uint8))
                masked_frame = frame * (mask / 255)
                gray = cv2.cvtColor(np.uint8(masked_frame), cv2.COLOR_BGR2GRAY)
                inners.append(get_inner(gray, i))
            if animation_analyzer is True:
                blobs.append(extract_blob(frame, i, channel=channel))

    return (contours, centers, heights, inners, blobs)


def generate_patternimage(frame, outlines, inners=None, std=0):
    if inners is not None:
        background_inners = np.zeros_like(frame)
        background_outers = np.zeros_like(frame)

    background_outlines = np.zeros_like(frame)

    if std > 0:
        backgrounds_std = []

    (y_bt, y_tp, x_lf, x_rt) = crop_frame(frame, outlines)

    length = len(outlines)
    p_size = int(max(abs(y_bt - y_tp), abs(x_lf - x_rt)) / 150 + 1)

    for n, outline in enumerate(outlines):
        if outline is not None:
            if std > 0:
                background_std = np.zeros_like(frame)
                if inners is not None:
                    cv2.drawContours(background_std, inners[n], -1, (255, 255, 255), -1)
                    backgrounds_std.append(background_std)

            if n < length / 4:
                d = n * int((255 * 4 / length))
                cv2.drawContours(background_outlines, [outline], 0, (255, d, 0), p_size)
                if inners is not None:
                    cv2.drawContours(
                        background_inners, inners[n], -1, (255, d, 0), p_size
                    )
            elif n < length / 2:
                d = int((n - length / 4) * (255 * 4 / length))
                cv2.drawContours(
                    background_outlines, [outline], 0, (255, 255, d), p_size
                )
                if inners is not None:
                    cv2.drawContours(
                        background_inners, inners[n], -1, (255, 255, d), p_size
                    )
            elif n < 3 * length / 4:
                d = int((n - length / 2) * (255 * 4 / length))
                cv2.drawContours(
                    background_outlines, [outline], 0, (255, 255 - d, 255), p_size
                )
                if inners is not None:
                    cv2.drawContours(
                        background_inners, inners[n], -1, (255, 255 - d, 255), p_size
                    )
            else:
                d = int((n - 3 * length / 4) * (255 * 4 / length))
                cv2.drawContours(
                    background_outlines, [outline], 0, (255 - d, 0, 255), p_size
                )
                if inners is not None:
                    cv2.drawContours(
                        background_inners, inners[n], -1, (255 - d, 0, 255), p_size
                    )

            if inners is not None:
                cv2.drawContours(
                    background_outers, [outline], 0, (255, 255, 255), int(2 * p_size)
                )

    outlines_image = background_outlines[y_bt:y_tp, x_lf:x_rt]

    if inners is not None:
        inners_image = background_inners[y_bt:y_tp, x_lf:x_rt]
        outers_image = background_outers[y_bt:y_tp, x_lf:x_rt]
        inners_image = cv2.subtract(inners_image, outers_image)

    if std > 0:
        backgrounds_std = np.array(backgrounds_std, dtype="float32")
        std_images = backgrounds_std[:, y_bt:y_tp, x_lf:x_rt]
        std_image = std_images.std(0)

        inners_image[std_image < std] = 0

    if inners is not None:
        pattern_image = cv2.add(inners_image, outlines_image)
    else:
        pattern_image = outlines_image

    return pattern_image


def generate_patternimage_all(
    frame, y_bt, y_tp, x_lf, x_rt, outlines_list, inners_list, std=0
):
    inners_length = len(inners_list[0])

    if inners_length > 0:
        background_inners = np.zeros_like(frame)
        background_outers = np.zeros_like(frame)

    background_outlines = np.zeros_like(frame)

    if std > 0:
        backgrounds_std = []

    length = len(outlines_list)
    p_size = int(max(abs(y_bt - y_tp), abs(x_lf - x_rt)) / 150 + 1)

    for n, outlines in enumerate(outlines_list):
        if std > 0:
            background_std = np.zeros_like(frame)
            if inners_length > 0:
                for inners in inners_list[n]:
                    cv2.drawContours(background_std, inners, -1, (255, 255, 255), -1)
                backgrounds_std.append(background_std)

        if n < length / 4:
            d = n * int((255 * 4 / length))
            cv2.drawContours(background_outlines, outlines, -1, (255, d, 0), p_size)
            if inners_length > 0:
                for inners in inners_list[n]:
                    cv2.drawContours(background_inners, inners, -1, (255, d, 0), p_size)
        elif n < length / 2:
            d = int((n - length / 4) * (255 * 4 / length))
            cv2.drawContours(background_outlines, outlines, -1, (255, 255, d), p_size)
            if inners_length > 0:
                for inners in inners_list[n]:
                    cv2.drawContours(
                        background_inners, inners, -1, (255, 255, d), p_size
                    )
        elif n < 3 * length / 4:
            d = int((n - length / 2) * (255 * 4 / length))
            cv2.drawContours(
                background_outlines, outlines, -1, (255, 255 - d, 255), p_size
            )
            if inners_length > 0:
                for inners in inners_list[n]:
                    cv2.drawContours(
                        background_inners, inners, -1, (255, 255 - d, 255), p_size
                    )
        else:
            d = int((n - 3 * length / 4) * (255 * 4 / length))
            cv2.drawContours(
                background_outlines, outlines, -1, (255 - d, 0, 255), p_size
            )
            if inners_length > 0:
                for inners in inners_list[n]:
                    cv2.drawContours(
                        background_inners, inners, -1, (255 - d, 0, 255), p_size
                    )

        if inners_length > 0:
            cv2.drawContours(
                background_outers, outlines, -1, (255, 255, 255), int(2 * p_size)
            )

    outlines_image = background_outlines[y_bt:y_tp, x_lf:x_rt]

    if inners_length > 0:
        inners_image = background_inners[y_bt:y_tp, x_lf:x_rt]
        outers_image = background_outers[y_bt:y_tp, x_lf:x_rt]
        inners_image = cv2.subtract(inners_image, outers_image)

    if std > 0:
        backgrounds_std = np.array(backgrounds_std, dtype="float32")
        std_images = backgrounds_std[:, y_bt:y_tp, x_lf:x_rt]
        std_image = std_images.std(0)

        inners_image[std_image < std] = 0

    if inners_length > 0:
        pattern_image = cv2.add(inners_image, outlines_image)
    else:
        pattern_image = outlines_image

    return pattern_image


def generate_patternimage_interact(
    frame, outlines, other_outlines, inners=None, other_inners=None, std=0
):
    total_outlines = functools.reduce(operator.iconcat, other_outlines, [])
    total_outlines += outlines
    (y_bt, y_tp, x_lf, x_rt) = crop_frame(frame, total_outlines)

    if inners is not None:
        background_inners = np.zeros_like(frame)
        background_outers = np.zeros_like(frame)

    background_outlines = np.zeros_like(frame)

    if std > 0:
        backgrounds_std = []

    length = len(outlines)
    p_size = int(max(abs(y_bt - y_tp), abs(x_lf - x_rt)) / 150 + 1)

    for n, outline in enumerate(outlines):
        other_outline = other_outlines[n]
        if len(other_outline) > 0:
            if other_outline[0] is not None:
                cv2.drawContours(
                    background_outlines, other_outline, -1, (150, 150, 150), p_size
                )

        if outline is not None:
            if inners is not None:
                inner = inners[n]
                other_inner = functools.reduce(operator.iconcat, other_inners[n], [])
                if other_inner is not None:
                    cv2.drawContours(
                        background_inners, other_inner, -1, (150, 150, 150), p_size
                    )
                if std > 0:
                    background_std = np.zeros_like(frame)
                    if inner is not None:
                        cv2.drawContours(background_std, inner, -1, (255, 255, 255), -1)
                    if other_inner is not None:
                        cv2.drawContours(
                            background_std, other_inner, -1, (255, 255, 255), -1
                        )
                    backgrounds_std.append(background_std)
            else:
                inner = None

            if n < length / 4:
                d = n * int((255 * 4 / length))
                cv2.drawContours(background_outlines, [outline], 0, (255, d, 0), p_size)
                if inner is not None:
                    cv2.drawContours(background_inners, inner, -1, (255, d, 0), p_size)
            elif n < length / 2:
                d = int((n - length / 4) * (255 * 4 / length))
                cv2.drawContours(
                    background_outlines, [outline], 0, (255, 255, d), p_size
                )
                if inner is not None:
                    cv2.drawContours(
                        background_inners, inner, -1, (255, 255, d), p_size
                    )
            elif n < 3 * length / 4:
                d = int((n - length / 2) * (255 * 4 / length))
                cv2.drawContours(
                    background_outlines, [outline], 0, (255, 255 - d, 255), p_size
                )
                if inner is not None:
                    cv2.drawContours(
                        background_inners, inner, -1, (255, 255 - d, 255), p_size
                    )
            else:
                d = int((n - 3 * length / 4) * (255 * 4 / length))
                cv2.drawContours(
                    background_outlines, [outline], 0, (255 - d, 0, 255), p_size
                )
                if inner is not None:
                    cv2.drawContours(
                        background_inners, inner, -1, (255 - d, 0, 255), p_size
                    )

            if inners is not None:
                cv2.drawContours(
                    background_outers, [outline], 0, (255, 255, 255), int(2 * p_size)
                )
                if len(other_outline) > 0:
                    if other_outline[0] is not None:
                        cv2.drawContours(
                            background_outers,
                            other_outline,
                            -1,
                            (150, 150, 150),
                            int(2 * p_size),
                        )

    outlines_image = background_outlines[y_bt:y_tp, x_lf:x_rt]

    if inners is not None:
        inners_image = background_inners[y_bt:y_tp, x_lf:x_rt]
        outers_image = background_outers[y_bt:y_tp, x_lf:x_rt]
        inners_image = cv2.subtract(inners_image, outers_image)

    if std > 0:
        backgrounds_std = np.array(backgrounds_std, dtype="float32")
        std_images = backgrounds_std[:, y_bt:y_tp, x_lf:x_rt]
        std_image = std_images.std(0)

        inners_image[std_image < std] = 0

    if inners is not None:
        pattern_image = cv2.add(inners_image, outlines_image)
    else:
        pattern_image = outlines_image

    return pattern_image


def plot_evnets(
    result_path,
    event_probability,
    time_points,
    names_and_colors,
    behavior_to_include,
    width=0,
    height=0,
):
    print("Exporting the raster plot for this analysis batch...")
    print(datetime.datetime.now())

    if width == 0 or height == 0:
        time_length = len(time_points)
        if time_length > 30000:
            width = round(time_length / 3000) + 1
            x_intvl = 3000
        elif time_length > 3000:
            width = round(time_length / 300) + 1
            x_intvl = 300
        else:
            width = round(time_length / 30) + 1
            x_intvl = 30
        height = round(len(event_probability) / 4) + 1
        if height < 3:
            height = 3
        figure, ax = plt.subplots(figsize=(width, height))
        if height <= 5:
            figure.subplots_adjust(bottom=0.25)

    for behavior_name in behavior_to_include:
        all_data = []
        masks = []

        for i in event_probability:
            data = []
            mask = []
            for n in range(len(event_probability[i])):
                if event_probability[i][n][0] == behavior_name:
                    data.append(event_probability[i][n][1])
                    mask.append(0)
                else:
                    data.append(-1)
                    mask.append(True)
            all_data.append(data)
            masks.append(mask)

        all_data = np.array(all_data)
        masks = np.array(masks)
        dataframe = pd.DataFrame(
            all_data, columns=[float("{:.1f}".format(i)) for i in time_points]
        )

        heatmap = sb.heatmap(
            dataframe,
            mask=masks,
            xticklabels=x_intvl,
            cmap=LinearSegmentedColormap.from_list("", names_and_colors[behavior_name]),
            cbar=False,
            vmin=0,
            vmax=1,
        )
        heatmap.set_xticklabels(heatmap.get_xticklabels(), rotation=90)
        # no ticks
        # ax.tick_params(axis='both',which='both',length=0)

    plt.savefig(os.path.join(result_path, "behaviors_plot.png"))

    for behavior_name in behavior_to_include:
        colorbar_fig = plt.figure(figsize=(5, 1))
        ax = colorbar_fig.add_axes([0, 1, 1, 1])
        colorbar = ColorbarBase(
            ax,
            orientation="horizontal",
            cmap=LinearSegmentedColormap.from_list("", names_and_colors[behavior_name]),
            norm=Normalize(vmin=0, vmax=1),
            ticks=[],
        )
        colorbar.outline.set_linewidth(0)

        plt.savefig(
            os.path.join(result_path, behavior_name + "_colorbar.png"),
            bbox_inches="tight",
        )
        plt.close()

    plt.close("all")

    print("The raster plot stored in: " + str(result_path))


def extract_frames(
    path_to_video, out_path, framewidth=None, start_t=0, duration=0, skip_redundant=1000
):
    capture = cv2.VideoCapture(path_to_video)
    fps = round(capture.get(cv2.CAP_PROP_FPS))
    full_duration = capture.get(cv2.CAP_PROP_FRAME_COUNT) / fps
    video_name = os.path.splitext(os.path.basename(path_to_video))[0]

    if start_t >= full_duration:
        print("The beginning time is later than the end of the video!")
        print("Will use the beginning of the video as the beginning time!")
        start_t = 0
    if duration <= 0:
        duration = full_duration
    end_t = start_t + duration

    frame_count = 1
    frame_count_generate = 0

    while True:
        retval, frame = capture.read()
        t = (frame_count) / fps
        if frame is None:
            break
        if t >= end_t:
            break

        if t >= start_t:
            if frame_count_generate % skip_redundant == 0:
                if framewidth is not None:
                    frameheight = int(frame.shape[0] * framewidth / frame.shape[1])
                    frame = cv2.resize(
                        frame, (framewidth, frameheight), interpolation=cv2.INTER_AREA
                    )

                cv2.imwrite(
                    os.path.join(
                        out_path, video_name + "_" + str(frame_count_generate) + ".jpg"
                    ),
                    frame,
                )

            frame_count_generate += 1

        frame_count += 1

    capture.release()

    print("The image examples stored in: " + out_path)


def preprocess_video(
    path_to_video: str,
    out_folder: str,
    framewidth: int | None,
    trim_video: bool = False,
    time_windows: list[Tuple[float, float]] = [(0, 10)],
    enhance_contrast: bool = True,
    contrast: float = 1.0,
    crop_frame: bool = True,
    left: int = 0,
    right: int = 0,
    top: int = 0,
    bottom: int = 0,
    fps_reduction_factor: float = 1.0,
):
    """
    Preprocess the given video for faster model training.

    Args:
        path_to_video: The path to the video to process.
        out_folder: The folder in which to store the processed video.
        framewidth: The width of the frame after resizing.
        trim_video: Whether or not to apply the given time windows.
        time_windows: A list of lists, where each inner list is of format
            [start, end] and indicates the start time and end time of a
            subsection of the video to include in preprocessing.
        enhance_contrast: Whether or not to enhance the contrast of the video.
            NOTE: The current implementation doesn't actually increase the
            contrast -- instead, it increases the brightness. This feature will
            need to be rewritten at some point.
        contrast: The factor by which to increase the brightness (see NOTE
            above for explanation).
        crop_frame: Whether or not to crop the video.
        left: The left coordinate to crop at.
        right: The right coordinate to crop at.
        top: The top coordinate to crop at.
        bottom: The bottom coordinate to crop at.
        fps_reduction_factor: The factor by which to reduce the FPS of the
            video. For example, if the video is originally at 60 FPS and
            fps_reduction_factor is 4, then the new video will be scaled down
            to 15 FPS while maintaining the same duration.

    Returns:
        None. The processed videos will be stored in `out_folder`.

    Raises:
        None
    """

    # Get video and metadata
    capture = cv2.VideoCapture(path_to_video)
    name = os.path.basename(path_to_video).split(".")[0]
    fps = round(capture.get(cv2.CAP_PROP_FPS))
    num_frames = int(capture.get(cv2.CAP_PROP_FRAME_COUNT))
    width = capture.get(cv2.CAP_PROP_FRAME_WIDTH)
    height = capture.get(cv2.CAP_PROP_FRAME_HEIGHT)

    # Calculate width and height if necessary
    if framewidth is not None:
        height = framewidth * height / width
        width = framewidth

    # Add video windows to filename if necessary
    added_name = ""
    if trim_video is True:
        for start, end in time_windows:
            added_name += "_" + str(start) + "-" + str(end)

    # Create VideoWriter with given parameters
    writer = cv2.VideoWriter(
        filename=os.path.join(out_folder, name + added_name + "_processed.avi"),
        fourcc=cv2.VideoWriter_fourcc(*"MJPG"),  # type: ignore
        fps=fps / fps_reduction_factor,
        frameSize=(int(width), int(height)),
        isColor=True,
    )

    # Get the indices of all dropped frames
    dropped_frames = get_dropped_frames(num_frames, fps_reduction_factor)

    frame_count = 0
    while True:
        frame_count += 1

        ret, frame = capture.read()
        if frame is None:
            break

        if frame_count - 1 in dropped_frames:
            continue

        # Apply preprocessing to the frame
        if framewidth is not None:
            frame = cv2.resize(
                frame,
                (int(width), int(height)),
                interpolation=cv2.INTER_AREA,
            )
        if crop_frame is True:
            frame = frame[top:bottom, left:right, :]
        if enhance_contrast is True:
            frame = frame * contrast  # type: ignore
            frame[frame > 255] = 255

        # Add frame to video file
        frame = np.uint8(frame)  # type: ignore
        if trim_video is True:
            t = frame_count / fps
            for start, end in time_windows:
                if start <= t <= end:
                    writer.write(frame)  # type: ignore
        else:
            writer.write(frame)  # type: ignore

    writer.release()
    capture.release()

    print("The processed video(s) stored in: " + out_folder)


def get_dropped_frames(n: int, reduction_factor: float) -> list[int]:
    """
    Return a list of indices of frames to be dropped after an FPS reduction

    Args:
        n: The number of frames in the original video.
        reduction_factor: The FPS reduction factor, where the new FPS is
            calculated using new_fps = old_fps / reduction_factor.

    Returns:
        A list of indices of frames to be dropped.

    Raises:
        None
    """
    if n <= 1.0:
        return []
    num_dropped_frames = int(n * (1 - 1 / reduction_factor))
    block_size = n / (n * (1 - 1 / reduction_factor) - 1)
    return [round(block_size * i) for i in range(num_dropped_frames)]



"""
Copyright (C)
This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with this program. If not, see https://tldrlegal.com/license/gnu-general-public-license-v3-(gpl-3)#fulltext. 

For license issues, please contact:

Dr. Bing Ye
Life Sciences Institute
University of Michigan
210 Washtenaw Avenue, Room 5403
Ann Arbor, MI 48109-2216
USA

Email: bingye@umich.edu
"""


import os

import wx

from LabGym.analyzebehaviorsdetector import (
    delete_detector,
    get_animal_names,
    get_annotation_class_names,
    get_detector_names,
    test_detector,
    train_detector,
)
from LabGym.gui.utils import WX_VIDEO_WILDCARD, LabGymWindow
from LabGym.tools import extract_frames


class GenerateImageExamples(LabGymWindow):
    """Generate image examples to use to train a Detector."""

    def __init__(self):
        """Open the Generate Image Example frame."""
        super().__init__(title="Generate Image Examples", size=(1000, 330))
        self.path_to_videos = None
        self.result_path = None
        self.framewidth = None
        self.t = 0
        self.duration = 0
        self.skip_redundant = 1000

        self.text_inputvideos = self.module_text("None.")
        self.add_module(
            "Select the video(s) to generate\nimage examples",
            self.select_videos,
            "Select one or more videos. Common video formats (mp4, mov, avi, m4v, mkv, mpg, mpeg) are supported except wmv format.",
            self.text_inputvideos,
        )

        self.text_outputfolder = self.module_text("None.")
        self.add_module(
            "Select a folder to store the\ngenerated image examples",
            self.select_outpath,
            "The generated image examples (extracted frames) will be stored in this folder.",
            self.text_outputfolder,
        )

        self.text_startgenerate = self.module_text(
            "Default: at the beginning of the video(s)."
        )
        self.add_module(
            "Specify when generating image examples\nshould begin (unit: second)",
            self.specify_timing,
            "Enter a beginning time point for all videos",
            self.text_startgenerate,
        )

        self.text_duration = self.module_text(
            "Default: from the specified beginning time to the end of a video."
        )
        self.add_module(
            "Specify how long generating examples\nshould last (unit: second)",
            self.input_duration,
            "This duration will be used for all the videos.",
            self.text_duration,
        )

        self.text_skipredundant = self.module_text(
            "Default: generate an image example every 1000 frames."
        )
        self.add_module(
            "Specify how many frames to skip when\ngenerating two consecutive images",
            self.specify_redundant,
            "To increase the efficiency of training a Detector, you need to make the training images as diverse (look different) as possible. You can do this by setting an interval between the two consecutively extracted images.",
            self.text_skipredundant,
        )

        self.add_submit_button(
            label="Start to generate image examples",
            handler=self.generate_images,
            tool_tip="Press the button to start generating image examples.",
        )

        self.display_window()

    def select_videos(self, event):
        """Open dialogs to select videos to generate images from."""
        video_select_dialog = wx.FileDialog(
            self, "Select video(s)", "", "", WX_VIDEO_WILDCARD, style=wx.FD_MULTIPLE
        )
        if video_select_dialog.ShowModal() != wx.ID_OK:
            video_select_dialog.Destroy()
            return

        self.path_to_videos = video_select_dialog.GetPaths()
        video_select_dialog.Destroy()

        path = os.path.dirname(self.path_to_videos[0])
        resize_dialog = wx.MessageDialog(
            self,
            'Proportional resize the video frames?\nSelect "No" if dont know what it is.',
            "(Optional) resize the frames?",
            wx.YES_NO | wx.ICON_QUESTION,
        )

        if resize_dialog.ShowModal() != wx.ID_YES:
            self.framewidth = None
            self.text_inputvideos.SetLabel(
                f"Selected {len(self.path_to_videos)} video(s) in: {path} (original framesize)."
            )
            resize_dialog.Destroy()
            return

        resize_dialog.Destroy()

        frame_width_dialog = wx.NumberEntryDialog(
            self,
            "Enter the desired frame width",
            "The unit is pixel:",
            "Desired frame width",
            480,
            1,
            10000,
        )

        if frame_width_dialog.ShowModal() != wx.ID_OK:
            self.framewidth = None
            self.text_inputvideos.SetLabel(
                f"Selected {len(self.path_to_videos)} video(s) in: {path} (original framesize)."
            )
            frame_width_dialog.Destroy()
            return

        self.framewidth = max(int(frame_width_dialog.GetValue()), 10)
        self.text_inputvideos.SetLabel(
            f"Selected {len(self.path_to_videos)} video(s) in: {path} (proportionally resize framewidth to {self.framewidth})."
        )
        frame_width_dialog.Destroy()

    def select_outpath(self, event):
        """Select directory to store image examples."""
        dialog = wx.DirDialog(self, "Select a directory", "", style=wx.DD_DEFAULT_STYLE)
        if dialog.ShowModal() == wx.ID_OK:
            self.result_path = dialog.GetPath()
            self.text_outputfolder.SetLabel(
                f"Generate image examples in: {self.result_path}."
            )
        dialog.Destroy()

    def specify_timing(self, event):
        """Choose time point to start generating image examples."""
        dialog = wx.NumberEntryDialog(
            parent=self,
            message="Enter beginning time to generate examples",
            prompt="The unit is second:",
            caption="Beginning time to generate examples",
            value=0,
            min=0,
            max=100000000000000,
        )
        if dialog.ShowModal() == wx.ID_OK:
            self.t = max(float(dialog.GetValue()), 0)
            self.text_startgenerate.SetLabel(
                f"Start to generate image examples at the: {str(self.t)} second."
            )
        dialog.Destroy()

    def input_duration(self, event):
        """Choose duration to generate video examples."""
        dialog = wx.NumberEntryDialog(
            parent=self,
            message="Enter the duration for generating examples",
            prompt="The unit is second:",
            caption="Duration for generating examples",
            value=0,
            min=0,
            max=100000000000000,
        )
        if dialog.ShowModal() == wx.ID_OK:
            self.duration = int(dialog.GetValue())
            self.text_duration.SetLabel(
                f"The generation of image examples lasts for {self.duration} seconds."
            )
        dialog.Destroy()

    def specify_redundant(self, event):
        """Select number of frames to skip per example."""
        dialog = wx.NumberEntryDialog(
            self,
            "How many frames to skip?",
            "Enter a number:",
            "Interval for generating examples",
            15,
            0,
            100000000000000,
        )
        self.skip_redundant = (
            int(dialog.GetValue()) if dialog.ShowModal() == wx.ID_OK else 1000
        )
        self.text_skipredundant.SetLabel(
            f"Generate an image example every {self.skip_redundant} frames."
        )
        dialog.Destroy()

    def generate_images(self, event):
        """Confirm image example generation."""
        if self.path_to_videos is None or self.result_path is None:
            wx.MessageBox(
                "No input video(s) / output folder selected.",
                "Error",
                wx.OK | wx.ICON_ERROR,
            )
            return

        dialog = wx.MessageDialog(
            self,
            "Start to generate image examples?",
            "Start to generate examples?",
            wx.YES_NO | wx.ICON_QUESTION,
        )

        if dialog.ShowModal() == wx.ID_YES:
            print("Generating image examples...")
            for i in self.path_to_videos:
                extract_frames(
                    i,
                    self.result_path,
                    framewidth=self.framewidth,
                    start_t=self.t,
                    duration=self.duration,
                    skip_redundant=self.skip_redundant,
                )
            print("Image example generation completed!")
        dialog.Destroy()


class TrainDetectors(LabGymWindow):
    """Train Detectors using labeled image examples."""

    def __init__(self):
        super().__init__(title="Train Detectors", size=(1000, 280))
        self.path_to_trainingimages = None
        self.path_to_annotation = None
        self.inference_size = 320
        self.iteration_num = 200

        self.text_selectimages = self.module_text("None.")
        self.add_module(
            button_label="Select the folder containing\nall the training images",
            button_handler=self.select_images,
            tool_tip="The folder that stores all the training images.",
            text=self.text_selectimages,
        )

        self.text_selectannotation = self.module_text("None.")
        self.add_module(
            button_label="Select the *.json\nannotation file",
            button_handler=self.select_annotation,
            tool_tip="The .json file that stores the annotation for the training images. Make sure it is in “COCO instance segmentation” format.",
            text=self.text_selectannotation,
        )

        self.text_inferencingsize = self.module_text("Default: 480.")
        self.add_module(
            button_label="Specify the inferencing framesize\nfor the Detector to train",
            button_handler=self.input_inferencingsize,
            tool_tip="This number should be divisible by 32. It determines the speed-accuracy trade-off of Detector performance. Larger size means higher accuracy but slower speed. See Extended Guide for details.",
            text=self.text_inferencingsize,
        )

        self.text_iterations = self.module_text("Default: 200.")
        self.add_module(
            button_label="Specify the iteration number\nfor the Detector training",
            button_handler=self.input_iterations,
            tool_tip='More training iterations typically yield higher accuracy but take longer. A number between 200 ~ 2000 is good for most scenarios. For "Interactive advanced" mode, more iterations (>2000) may be needed. You may also increase the diversity and amount of training images for higher accuracy.',
            text=self.text_iterations,
        )

        self.add_submit_button(
            label="Train the Detector",
            handler=self.train_detector,
            tool_tip="You need to name the Detector to train. English letters, numbers, underscore “_”, or hyphen “-” are acceptable but do not use special characters such as “@” or “^”.",
        )

        self.display_window()

    def select_images(self, event):
        """Select labeled images."""
        dialog = wx.DirDialog(self, "Select a directory", "", style=wx.DD_DEFAULT_STYLE)
        if dialog.ShowModal() == wx.ID_OK:
            self.path_to_trainingimages = dialog.GetPath()
            self.text_selectimages.SetLabel(
                f"Path to training images: {self.path_to_trainingimages}."
            )
        dialog.Destroy()

    def select_annotation(self, event):
        """Select COCO annotation file for images."""
        dialog = wx.FileDialog(
            self,
            "Select the annotation file (.json)",
            "",
            wildcard="Annotation File (*.json)|*.json",
            style=wx.FD_OPEN,
        )
        if dialog.ShowModal() == wx.ID_OK:
            self.path_to_annotation = dialog.GetPath()
            class_names = get_annotation_class_names(self.path_to_annotation)
            self.text_selectannotation.SetLabel(
                f"Animal/object categories in annotation file: {class_names}."
            )
        dialog.Destroy()

    @property
    def inference_size(self) -> int:
        """The inferencing frame size for the Detector."""
        return self._inference_size

    @inference_size.setter
    def inference_size(self, inference_size: int):
        """Set the inferencing frame size.

        Args:
            inference_size: The new inferencing frame size.

        Raises:
            ValueError: The frame size is not a multiple of 32.
        """
        if inference_size % 32 != 0:
            raise ValueError(f"{inference_size} is not a multiple of 32.")
        self._inference_size = inference_size

    def input_inferencingsize(self, event):
        """Enter inferencing frame size for the Detector."""
        dialog = wx.NumberEntryDialog(
            self,
            "Input the inferencing frame size\nof the Detector to train",
            "Enter a number:",
            "Divisible by 32",
            value=self.inference_size,
            min=32,
            max=2048,
        )
        if dialog.ShowModal() != wx.ID_OK:
            dialog.Destroy()
            return

        try:
            self.inference_size = int(dialog.GetValue())
            self.text_inferencingsize.SetLabel(
                "Inferencing frame size: " + str(self.inference_size) + "."
            )
        except ValueError as err:
            wx.MessageBox(
                str(err),
                "Error",
                wx.OK | wx.ICON_ERROR,
            )
        dialog.Destroy()

    def input_iterations(self, event):
        """Set the number of iterations for Detector training."""
        dialog = wx.NumberEntryDialog(
            self,
            "Input the iteration number\nfor the Detector training",
            "Enter a number:",
            "Iterations",
            value=self.iteration_num,
            min=1,
            max=10000,
        )
        if dialog.ShowModal() == wx.ID_OK:
            self.iteration_num = int(dialog.GetValue())
            self.text_iterations.SetLabel(
                f"Training iteration number: {self.iteration_num}."
            )
        dialog.Destroy()

    def train_detector(self, event):
        """Train a Detector with the given parameters."""
        if self.path_to_trainingimages is None or self.path_to_annotation is None:
            wx.MessageBox(
                "No training images / annotation file selected.",
                "Error",
                wx.OK | wx.ICON_ERROR,
            )
            return

        while True:
            dialog = wx.TextEntryDialog(
                self, "Enter a name for the Detector to train", "Detector name"
            )
            if dialog.ShowModal() != wx.ID_OK:
                break

            if dialog.GetValue().strip() == "":
                wx.MessageBox(
                    "Please enter a name.",
                    "Error",
                    wx.OK | wx.ICON_ERROR,
                )
                continue

            detector_name = dialog.GetValue()
            if detector_name in get_detector_names():
                wx.MessageBox(
                    f"The Detector {dialog.GetValue()} already exists!",
                    "Error",
                    wx.OK | wx.ICON_ERROR,
                )
                continue

            train_detector(
                self.path_to_annotation,
                self.path_to_trainingimages,
                detector_name,
                self.iteration_num,
                self.inference_size,
            )
            break


class TestDetectors(LabGymWindow):
    """Test Detectors using labeled image examples."""

    def __init__(self):
        super().__init__(title="Test Detectors", size=(1000, 280))
        self.path_to_testingimages = None
        self.path_to_annotation = None
        self.detector_name = None
        self.results_folder = None

        self.text_selectdetector = self.module_text("None.")
        self.add_module(
            "Select a Detector\nto test",
            self.select_detector,
            "The object / animal names in the ground-truth testing image dataset should match those in the selected Detector.",
            self.text_selectdetector,
        )

        self.text_selectimages = self.module_text("None.")
        self.add_module(
            "Select the folder containing\nall the testing images",
            self.select_images,
            "The folder that stores all the testing images.",
            self.text_selectimages,
        )

        self.text_selectannotation = self.module_text("None.")
        self.add_module(
            "Select the *.json\nannotation file",
            self.select_annotation,
            "The .json file that stores the annotation for the testing images. Make sure it is in “COCO instance segmentation” format.",
            self.text_selectannotation,
        )

        self.text_select_results_folder = self.module_text("None.")
        self.add_module(
            "Select the folder to\nstore testing results",
            self.select_results_folder,
            "The folder will stores the testing results.",
            self.text_select_results_folder,
        )

        testanddelete = wx.BoxSizer(wx.HORIZONTAL)
        button_test = wx.Button(self.panel, label="Test the Detector", size=(300, 40))
        button_test.Bind(wx.EVT_BUTTON, self.test_detector)
        wx.Button.SetToolTip(
            button_test,
            "Test the selected Detector on the annotated, ground-truth testing images.",
        )
        button_delete = wx.Button(self.panel, label="Delete a Detector", size=(300, 40))
        button_delete.Bind(wx.EVT_BUTTON, self.delete_detector)
        wx.Button.SetToolTip(
            button_delete,
            "Permanently delete a Detector. The deletion CANNOT be restored.",
        )
        testanddelete.Add(button_test, 0, wx.RIGHT, 5)
        testanddelete.Add(button_delete, 0, wx.LEFT, 5)
        self.boxsizer.Add(0, self.MODULE_TOP_MARGIN, 0)
        self.boxsizer.Add(
            testanddelete, 0, wx.RIGHT | wx.ALIGN_RIGHT, self.BOTTOM_MARGIN * 2
        )
        self.boxsizer.Add(0, self.BOTTOM_MARGIN, 0)

        self.display_window()

    def select_detector(self, event):
        """Select a detector to test."""
        dialog = wx.SingleChoiceDialog(
            self,
            message="Select a Detector to test",
            caption="Test a Detector",
            choices=get_detector_names(),
        )

        if dialog.ShowModal() == wx.ID_OK:
            self.detector_name = dialog.GetStringSelection()
            self.text_selectdetector.SetLabel(
                f"Selected: {self.detector_name} (animals/objects: {get_animal_names(self.detector_name)})."
            )
        dialog.Destroy()

    def select_images(self, event):
        """Select annotated testing images."""
        dialog = wx.DirDialog(self, "Select a directory", "", style=wx.DD_DEFAULT_STYLE)
        if dialog.ShowModal() == wx.ID_OK:
            self.path_to_testingimages = dialog.GetPath()
            self.text_selectimages.SetLabel(
                f"Path to testing images: {self.path_to_testingimages}."
            )
        dialog.Destroy()

    def select_annotation(self, event):
        """Select COCO annotation file."""
        dialog = wx.FileDialog(
            self,
            "Select the annotation file (.json)",
            "",
            wildcard="Annotation File (*.json)|*.json",
            style=wx.FD_OPEN,
        )
        if dialog.ShowModal() == wx.ID_OK:
            self.path_to_annotation = dialog.GetPath()
            class_names = get_annotation_class_names(self.path_to_annotation)
            self.text_selectannotation.SetLabel(
                f"Animal/object categories in annotation file: {class_names}."
            )
        dialog.Destroy()

    def select_results_folder(self, event):
        """Select folder to store testing results."""
        dialog = wx.DirDialog(self, "Select a directory", "", style=wx.DD_DEFAULT_STYLE)
        if dialog.ShowModal() == wx.ID_OK:
            self.results_folder = dialog.GetPath()
            self.text_select_results_folder.SetLabel(
                f"Testing results stored in: {self.results_folder}."
            )
        dialog.Destroy()

    def test_detector(self, event):
        """Test the selected detector."""
        if (
            self.detector_name is None
            or self.path_to_testingimages is None
            or self.path_to_annotation is None
            or self.results_folder is None
        ):
            wx.MessageBox(
                "No Detector / training images / annotation file / output path selected.",
                "Error",
                wx.OK | wx.ICON_ERROR,
            )
        else:
            test_detector(
                self.path_to_annotation,
                self.path_to_testingimages,
                self.detector_name,
                self.results_folder,
            )

    def delete_detector(self, event):
        """Delete a Detector."""
        dialog = wx.SingleChoiceDialog(
            self,
            message="Select a Detector to delete",
            caption="Delete a Detector",
            choices=get_detector_names().sort(),
        )
        if dialog.ShowModal() != wx.ID_OK:
            dialog.Destroy()
            return

        detector = dialog.GetStringSelection()
        confirm = wx.MessageDialog(
            self,
            f"Delete {detector}?",
            "CANNOT be restored!",
            wx.YES_NO | wx.ICON_QUESTION,
        )
        if confirm.ShowModal() == wx.ID_YES:
            delete_detector(detector)
        confirm.Destroy()



"""
Copyright (C)
This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with this program. If not, see https://tldrlegal.com/license/gnu-general-public-license-v3-(gpl-3)#fulltext. 

For license issues, please contact:

Dr. Bing Ye
Life Sciences Institute
University of Michigan
210 Washtenaw Avenue, Room 5403
Ann Arbor, MI 48109-2216
USA

Email: bingye@umich.edu
"""


import wx
import wx.lib.agw.hyperlink

from LabGym import __version__

from .analysis import AnalysisModule
from .preprocessing import PreprocessingModule
from .training import TrainingModule


class MainWindow(wx.Frame):
    """The main LabGym window."""

    def __init__(self, title):
        super().__init__(parent=None, title=title, size=(750, 440))
        panel = wx.Panel(self)
        boxsizer = wx.BoxSizer(wx.VERTICAL)

        # Welcome text
        self.welcome_text = wx.StaticText(
            panel,
            label="Welcome to LabGym!",
            style=wx.ALIGN_CENTER | wx.ST_ELLIPSIZE_END,
        )
        boxsizer.Add(0, 60, 0)
        boxsizer.Add(self.welcome_text, 0, wx.LEFT | wx.RIGHT | wx.EXPAND, 5)
        boxsizer.Add(0, 60, 0)

        # Developers text
        self.developers_text = wx.StaticText(
            panel,
            label="Developed by Yujia Hu, Rohan Satapathy, M. Victor Struman, Kelly Goss, Isabelle Baker\n\nBing Ye Lab, Life Sciences Institute, University of Michigan",
            style=wx.ALIGN_CENTER | wx.ST_ELLIPSIZE_END,
        )
        boxsizer.Add(self.developers_text, 0, wx.LEFT | wx.RIGHT | wx.EXPAND, 5)
        boxsizer.Add(0, 60, 0)

        # Links to GitHub and extended user guide
        links = wx.BoxSizer(wx.HORIZONTAL)
        homepage = wx.lib.agw.hyperlink.HyperLinkCtrl(
            panel, 0, "Home Page", URL="https://github.com/umyelab/LabGym"
        )
        userguide = wx.lib.agw.hyperlink.HyperLinkCtrl(
            panel,
            0,
            "Extended Guide",
            URL="https://github.com/yujiahu415/LabGym/blob/master/LabGym%20user%20guide_v2.2.pdf",
        )
        links.Add(homepage, 0, wx.LEFT | wx.EXPAND, 10)
        links.Add(userguide, 0, wx.LEFT | wx.RIGHT | wx.EXPAND, 10)
        boxsizer.Add(links, 0, wx.ALIGN_CENTER, 50)
        boxsizer.Add(0, 50, 0)

        # Main modules
        modules = wx.BoxSizer(wx.HORIZONTAL)
        preprocessing_button = wx.Button(
            panel, label="Preprocessing Module", size=(200, 40)
        )
        preprocessing_button.Bind(wx.EVT_BUTTON, self.open_preprocessing_module)
        wx.Button.SetToolTip(
            preprocessing_button,
            "Enhance video contrast / crop frames to exclude unnecessary region / trim videos to only keep necessary time windows.",
        )
        training_button = wx.Button(panel, label="Training Module", size=(200, 40))
        training_button.Bind(wx.EVT_BUTTON, self.open_training_module)
        wx.Button.SetToolTip(
            training_button,
            "Teach LabGym to recognize the animals / objects of your interest and identify their behaviors that are defined by you.",
        )
        analysis_button = wx.Button(panel, label="Analysis Module", size=(200, 40))
        analysis_button.Bind(wx.EVT_BUTTON, self.open_analysis_module)
        wx.Button.SetToolTip(
            analysis_button,
            "Use LabGym to track the animals / objects of your interest, identify and quantify their behaviors, and display the statistically significant findings.",
        )
        modules.Add(preprocessing_button, 0, wx.LEFT | wx.RIGHT | wx.EXPAND, 10)
        modules.Add(training_button, 0, wx.LEFT | wx.RIGHT | wx.EXPAND, 10)
        modules.Add(analysis_button, 0, wx.LEFT | wx.RIGHT | wx.EXPAND, 10)
        boxsizer.Add(modules, 0, wx.ALIGN_CENTER, 50)
        boxsizer.Add(0, 50, 0)

        panel.SetSizer(boxsizer)

        self.Centre()
        self.Show(True)

    def open_preprocessing_module(self, event):
        """Opens the preprocessing module."""
        PreprocessingModule()

    def open_training_module(self, event):
        """Opens the training module."""
        TrainingModule()

    def open_analysis_module(self, event):
        """Opens the analysis module."""
        AnalysisModule()


def gui():
    app = wx.App()
    MainWindow(f"LabGym v{__version__}")
    print("The user interface initialized!")
    app.MainLoop()


if __name__ == "__main__":
    gui()



"""
Copyright (C)
This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with this program. If not, see https://tldrlegal.com/license/gnu-general-public-license-v3-(gpl-3)#fulltext. 

For license issues, please contact:

Dr. Bing Ye
Life Sciences Institute
University of Michigan
210 Washtenaw Avenue, Room 5403
Ann Arbor, MI 48109-2216
USA

Email: bingye@umich.edu
"""

import os

import pandas as pd
import wx

from LabGym.minedata import data_mining


class MineResults(wx.Frame):
    def __init__(self):
        super().__init__(parent=None, title="Mine Results", size=(1000, 240))
        self.file_path = None
        self.result_path = None
        self.dataset = None
        self.paired = False
        self.control = None
        self.pval = 0.05
        self.file_names = None
        self.control_file_name = None

        self.display_window()

    def display_window(self):
        panel = wx.Panel(self)
        boxsizer = wx.BoxSizer(wx.VERTICAL)

        module_inputfolder = wx.BoxSizer(wx.HORIZONTAL)
        button_inputfolder = wx.Button(
            panel, label="Select the folder that stores\nthe data files", size=(300, 40)
        )
        button_inputfolder.Bind(wx.EVT_BUTTON, self.select_filepath)
        wx.Button.SetToolTip(
            button_inputfolder,
            "Put all folders that store analysis results (each folder contains one raster plot) for control / experimental groups into this folder.",
        )
        self.text_inputfolder = wx.StaticText(
            panel, label="None.", style=wx.ALIGN_LEFT | wx.ST_ELLIPSIZE_END
        )
        module_inputfolder.Add(
            button_inputfolder, 0, wx.LEFT | wx.RIGHT | wx.EXPAND, 10
        )
        module_inputfolder.Add(
            self.text_inputfolder, 0, wx.LEFT | wx.RIGHT | wx.EXPAND, 10
        )
        boxsizer.Add(0, 10, 0)
        boxsizer.Add(module_inputfolder, 0, wx.LEFT | wx.RIGHT | wx.EXPAND, 10)
        boxsizer.Add(0, 5, 0)

        module_selectcontrol = wx.BoxSizer(wx.HORIZONTAL)
        button_selectcontrol = wx.Button(
            panel, label="Select the\ncontrol group", size=(300, 40)
        )
        button_selectcontrol.Bind(wx.EVT_BUTTON, self.select_control)
        wx.Button.SetToolTip(
            button_selectcontrol,
            "For multiple-group comparison, you can select one group as control for post-hoc comparison. If no control is selected, post-hoc comparison will be performed between each pair of the two groups.",
        )
        self.text_selectcontrol = wx.StaticText(
            panel,
            label="Default: no control group.",
            style=wx.ALIGN_LEFT | wx.ST_ELLIPSIZE_END,
        )
        module_selectcontrol.Add(
            button_selectcontrol, 0, wx.LEFT | wx.RIGHT | wx.EXPAND, 10
        )
        module_selectcontrol.Add(
            self.text_selectcontrol, 0, wx.LEFT | wx.RIGHT | wx.EXPAND, 10
        )
        boxsizer.Add(module_selectcontrol, 0, wx.LEFT | wx.RIGHT | wx.EXPAND, 10)
        boxsizer.Add(0, 5, 0)

        module_outputfolder = wx.BoxSizer(wx.HORIZONTAL)
        button_outputfolder = wx.Button(
            panel,
            label="Select the folder to store\nthe data mining results",
            size=(300, 40),
        )
        button_outputfolder.Bind(wx.EVT_BUTTON, self.select_result_path)
        wx.Button.SetToolTip(
            button_outputfolder,
            "A spreadsheet containing all data mining results will be stored in this folder.",
        )
        self.text_outputfolder = wx.StaticText(
            panel, label="None.", style=wx.ALIGN_LEFT | wx.ST_ELLIPSIZE_END
        )
        module_outputfolder.Add(
            button_outputfolder, 0, wx.LEFT | wx.RIGHT | wx.EXPAND, 10
        )
        module_outputfolder.Add(
            self.text_outputfolder, 0, wx.LEFT | wx.RIGHT | wx.EXPAND, 10
        )
        boxsizer.Add(module_outputfolder, 0, wx.LEFT | wx.RIGHT | wx.EXPAND, 10)
        boxsizer.Add(0, 5, 0)

        button_minedata = wx.Button(panel, label="Start to mine data", size=(300, 40))
        button_minedata.Bind(wx.EVT_BUTTON, self.mine_data)
        wx.Button.SetToolTip(
            button_minedata,
            "Parametric / non-parametric tests will be automatically selected according to the data distribution, to compare the mean / median of different groups. See Extended Guide for details.",
        )
        boxsizer.Add(0, 5, 0)
        boxsizer.Add(button_minedata, 0, wx.RIGHT | wx.ALIGN_RIGHT, 90)
        boxsizer.Add(0, 10, 0)

        panel.SetSizer(boxsizer)

        self.Centre()
        self.Show(True)

    def select_filepath(self, event):
        dialog = wx.MessageDialog(
            self, "Is the data paired?", "Paired data?", wx.YES_NO | wx.ICON_QUESTION
        )
        if dialog.ShowModal() == wx.ID_YES:
            self.paired = True
        else:
            self.paired = False
        dialog.Destroy()

        dialog = wx.DirDialog(self, "Select a directory", "", style=wx.DD_DEFAULT_STYLE)
        if dialog.ShowModal() == wx.ID_OK:
            self.file_path = dialog.GetPath()
            if self.paired is True:
                self.text_inputfolder.SetLabel(
                    "Paired input data is in: " + self.file_path + "."
                )
            else:
                self.text_inputfolder.SetLabel(
                    "Unpaired input data is in: " + self.file_path + "."
                )
        dialog.Destroy()

    def select_control(self, event):
        dialog = wx.SingleChoiceDialog(
            self,
            "Select the folder for the control group.",
            "Ignore if you wish to compare all groups",
            choices=[
                i
                for i in os.listdir(self.file_path)
                if os.path.isdir(os.path.join(self.file_path, i))
            ],
            style=wx.DD_DEFAULT_STYLE,
        )
        if dialog.ShowModal() == wx.ID_OK:
            control_path = dialog.GetStringSelection()
            self.text_selectcontrol.SetLabel(
                "The control group is: " + control_path + "."
            )
            self.control = self.read_folder(os.path.join(self.file_path, control_path))
            self.control_file_name = os.path.split(control_path)[1]
        else:
            self.control = None
            self.text_selectcontrol.SetLabel("No control group.")
        dialog.Destroy()

    def select_result_path(self, event):
        dialog = wx.DirDialog(self, "Select a directory", "", style=wx.DD_DEFAULT_STYLE)
        if dialog.ShowModal() == wx.ID_OK:
            self.result_path = dialog.GetPath()
            self.text_outputfolder.SetLabel(
                "Mining results are in: " + self.result_path + "."
            )
        dialog.Destroy()

    def read_folder(self, folder):
        folder = folder.replace("\\", "/")
        filelist = {}
        df = {}
        for i in os.listdir(folder):
            if (
                i.endswith("_summary.xlsx")
                or i.endswith("_summary.xls")
                or i.endswith("_summary.XLS")
            ):
                behavior_name = i.split("_")[-2]
                filelist[behavior_name] = os.path.join(folder, i)
        if len(filelist) == 0:
            print('No "_summary.xlsx" excel file found!')
        else:
            for behavior_name in filelist:
                dataset = pd.read_excel(r"" + filelist[behavior_name])
                dataset = dataset.drop(columns=["ID/parameter"])
                df[behavior_name] = dataset
        return df

    def read_all_folders(self):
        data = []
        filenames = []
        for file in os.listdir(self.file_path):
            subfolder = os.path.join(self.file_path, file)
            if os.path.isdir(subfolder):
                folder_data = self.read_folder(subfolder)
                if len(folder_data) > 0:
                    data.append(folder_data)
                    filenames.append(os.path.split(subfolder)[1])
                else:
                    print("Improper file / file structure in: " + subfolder + ".")
        self.dataset = data
        self.file_names = filenames

    def control_organization(self):
        if self.control == None:
            return
        del_idx = self.file_names.index(self.control_file_name)
        self.dataset.pop(del_idx)
        self.file_names.insert(0, self.file_names.pop(del_idx))

    def mine_data(self, event):
        if self.file_path is None or self.result_path is None:
            wx.MessageBox(
                "No input / output folder selected.", "Error", wx.OK | wx.ICON_ERROR
            )

        else:
            dialog = wx.TextEntryDialog(
                self,
                "Enter a p-value to determine statistical significance",
                "Default p-value is 0.05",
                "0.05",
            )
            if dialog.ShowModal() == wx.ID_OK:
                self.pval = float(dialog.GetValue())
            dialog.Destroy()

            print("Start to mine analysis results...")

            self.read_all_folders()
            self.control_organization()
            DM = data_mining(
                self.dataset,
                self.control,
                self.paired,
                self.result_path,
                self.pval,
                self.file_names,
            )
            DM.statistical_analysis()



"""
Copyright (C)
This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with this program. If not, see https://tldrlegal.com/license/gnu-general-public-license-v3-(gpl-3)#fulltext. 

For license issues, please contact:

Dr. Bing Ye
Life Sciences Institute
University of Michigan
210 Washtenaw Avenue, Room 5403
Ann Arbor, MI 48109-2216
USA

Email: bingye@umich.edu
"""


import os
from pathlib import Path
from typing import Tuple

import cv2
import numpy as np
import wx

from LabGym.tools import preprocess_video

from .utils import WX_VIDEO_WILDCARD, LabGymWindow


class PreprocessingModule(LabGymWindow):
    """Contains functions related to preprocessing videos for analysis."""

    def __init__(self):
        super().__init__(title="Preprocess Videos", size=(1000, 370))
        self.video_paths = None
        self.framewidth = None
        self.result_path = None
        self.trim_video = False
        self.time_windows = []
        self.enhance_contrast = False
        self.contrast = 1.0
        self.crop_frame = False
        self.left = 0
        self.right = 0
        self.top = 0
        self.bottom = 0
        self.decode_t = False
        self.fps_reduction_factor = 1.0

        # Video selection module
        self.text_inputvideos = self.module_text("None.")
        self.add_module(
            button_label="Select the video(s)\nfor preprocessing",
            button_handler=self.select_videos,
            tool_tip="Select one or more videos. Common video formats (mp4, mov, avi, m4v, mkv, mpg, mpeg) are supported except wmv format.",
            text=self.text_inputvideos,
        )

        # Output folder selection module
        self.text_outputfolder = self.module_text("None.")
        self.add_module(
            button_label="Select a folder to store\nthe processed videos",
            button_handler=self.select_outpath,
            tool_tip="Will create a subfolder for each video in the selected folder. Each subfolder is named after the file name of the video.",
            text=self.text_outputfolder,
        )

        # Video trimming module
        self.text_duration = self.module_text("Default: not to trim a video.")
        self.add_module(
            button_label="Specify whether to enter time windows\nto form a trimmed video",
            button_handler=self.input_duration,
            tool_tip='If "Yes", specify time windows by format: starttime1-endtime1,starttime2-endtime2,...to form the new, trimmed videos. See Extended Guide how to set different time windows for different videos.',
            text=self.text_duration,
        )

        # Frame crop module
        self.text_cropframe = self.module_text("Default: not to crop frames.")
        self.add_module(
            button_label="Specify whether to crop\nthe video frames",
            button_handler=self.crop_frames,
            tool_tip="Cropping frames to exclude unnecessary areas can increase the analysis efficiency. You need to specify the 4 corner points of the cropping window. This cropping window will be applied for all videos selected.",
            text=self.text_cropframe,
        )

        # Contrast module
        # TODO: This feature increases brightness, not contrast (see tools.preprocess_video())
        self.text_enhancecontrast = self.module_text(
            "Default: not to enhance contrast."
        )
        self.add_module(
            button_label="Specify whether to enhance\nthe contrast in videos",
            button_handler=self.enhance_contrasts,
            tool_tip="Enhancing video contrast will increase the detection accuracy especially when the detection method is background subtraction based. Enter a contrast value to see whether it is good to apply or re-enter it.",
            text=self.text_enhancecontrast,
        )

        # FPS reduction module
        self.text_fps = self.module_text("Default: original FPS")
        self.add_module(
            button_label="Specify whether to reduce\nthe video FPS",
            button_handler=self.reduce_fps,
            tool_tip="Reducing video FPS will decrease model training time",
            text=self.text_fps,
        )

        # Start button
        self.add_submit_button(
            label="Start to preprocess the videos",
            handler=self.preprocess_videos,
            tool_tip="Preprocess each selected video.",
        )

        self.display_window()

    def select_videos(self, event):
        """Opens file selection dialog to select videos to preprocess."""

        # Select videos
        video_selection_dialog = wx.FileDialog(
            parent=self,
            message="Select video(s)",
            wildcard=WX_VIDEO_WILDCARD,
            style=wx.FD_MULTIPLE,
        )

        if video_selection_dialog.ShowModal() != wx.ID_OK:
            video_selection_dialog.Destroy()
            return

        self.video_paths = video_selection_dialog.GetPaths()
        video_selection_dialog.Destroy()
        self.video_paths.sort()
        video_folder = os.path.dirname(self.video_paths[0])
        label_text = f"Selected {len(self.video_paths)} video(s) in: {video_folder}"

        # Ask whether or not to resize the videos
        resize_dialog = wx.MessageDialog(
            parent=self,
            message="Proportional resize the video frames?",
            caption="(Optional) resize the frames?",
            style=wx.YES_NO | wx.ICON_QUESTION,
        )

        if resize_dialog.ShowModal() != wx.ID_YES:
            self.framewidth = None
            self.text_inputvideos.SetLabel(label_text + " (original framesize).")
            resize_dialog.Destroy()
            return

        # Ask for the new frame width
        frame_width_dialog = wx.NumberEntryDialog(
            parent=self,
            message="Enter the desired frame width",
            prompt="The unit is pixel:",
            caption="Desired frame width",
            value=480,
            min=10,
            max=10000,
        )

        if frame_width_dialog.ShowModal() == wx.ID_OK:
            self.framewidth = int(frame_width_dialog.GetValue())
            label_text += f" (proportionally resize framewidth to {self.framewidth})."
        else:
            self.framewidth = None
            label_text += " (original framesize)."

        frame_width_dialog.Destroy()
        resize_dialog.Destroy()
        self.text_inputvideos.SetLabel(label_text)

    def select_outpath(self, event):
        """Opens folder selection dialog for storing processed video."""
        dialog = wx.DirDialog(
            parent=self, message="Select a directory", style=wx.DD_DEFAULT_STYLE
        )
        if dialog.ShowModal() == wx.ID_OK:
            self.result_path = dialog.GetPath()
            self.text_outputfolder.SetLabel(
                f"Processed videos will be in: {self.result_path}."
            )
        dialog.Destroy()

    def input_duration(self, event):
        """Opens dialogs to configure input time windows."""

        # Ask whether or not to trim a video
        trim_dialog = wx.MessageDialog(
            self,
            "Whether to trim a video?",
            "Trim videos?",
            wx.YES_NO | wx.ICON_QUESTION,
        )
        self.trim_video = trim_dialog.ShowModal() == wx.ID_YES
        trim_dialog.Destroy()

        if not self.trim_video:
            return

        # Ask for method for time windows
        methods = [
            'Decode from filenames: "_stt_" and "_edt_"',
            "Enter time points",
        ]
        method_dialog = wx.SingleChoiceDialog(
            self,
            message="Specify the time windows for trimming videos",
            caption="Time windows for trimming videos",
            choices=methods,
        )

        if method_dialog.ShowModal() != wx.ID_OK:
            method_dialog.Destroy()
            return

        self.decode_t = False
        method = method_dialog.GetStringSelection()
        if method == methods[0]:  # Use st<start> and ed<end> tags
            self.decode_t = True
            self.text_duration.SetLabel('Decode from filenames: "_stt_" and "_edt_"')
            method_dialog.Destroy()
            return

        # Manual entry for time windows
        windows_dialog = wx.TextEntryDialog(
            self,
            "Format: starttime1-endtime1,starttime2-endtime2,...",
            "Enter the time windows (in seconds)",
        )

        if windows_dialog.ShowModal() != wx.ID_OK:
            self.trim_video = False
            self.text_duration.SetLabel("Not to trim the videos.")
            windows_dialog.Destroy()
            method_dialog.Destroy()
            return

        try:
            window_input: str = windows_dialog.GetValue()
            self.time_windows = self.parse_time_window_str(window_input)
            self.text_duration.SetLabel(
                f"The time windows to form the new, trimmed video: {self.time_windows}."
            )
        except ValueError as err:
            self.trim_video = False
            self.text_duration.SetLabel("Not to trim the videos.")
            wx.MessageBox(
                f"{str(err)} Please enter the time windows in correct format!",
                "Error",
                wx.OK | wx.ICON_ERROR,
            )
        windows_dialog.Destroy()
        method_dialog.Destroy()

    def parse_time_window_str(self, window_input: str) -> list[Tuple[float, float]]:
        """
        Return the time windows to trim the videos by given user input.

        The user should input the time windows in the format
            start-end, start-end, start-end...
        where start and end represent times in seconds such that start < end.

        Args:
            windows: The user's time window entry.

        Returns:
            A list of tuples (start, end) where start and end are floats that
            represent the start and end times in seconds of each time window.

        Raises:
            ValueError: There was an error parsing the time window string.
        """
        windows = []
        for window in window_input.split(","):
            w = [i.strip() for i in window.split("-")]
            if len(w) != 2:
                raise ValueError(f"'{window}' is an invalid time window.")
            try:
                start = float(w[0].strip())
            except ValueError:
                raise ValueError(f"'{w[0]}' is an invalid start time.")
            try:
                end = float(w[1])
            except ValueError:
                raise ValueError(f"'{w[1]}' is an invalid end time.")
            if not start <= end:
                raise ValueError("Start time must be less than end time.")
            windows.append((start, end))
        return windows

    def crop_frames(self, event):
        """Opens dialogs to configure frame cropping."""

        if self.video_paths is None:
            wx.MessageBox("No video selected.", "Error", wx.OK | wx.ICON_ERROR)
            return

        capture = cv2.VideoCapture(self.video_paths[0])
        _, frame = capture.read()
        capture.release()

        if self.framewidth is not None:
            frame = cv2.resize(
                frame,
                (
                    self.framewidth,
                    int(frame.shape[0] * self.framewidth / frame.shape[1]),
                ),
                interpolation=cv2.INTER_AREA,
            )

        canvas = np.copy(frame)
        height, width = frame.shape[:2]
        for y in range(0, height, 50):
            cv2.line(canvas, (0, y), (width, y), (255, 0, 255), 1)
            cv2.putText(
                canvas,
                str(y),
                (5, y + 15),
                cv2.FONT_HERSHEY_SIMPLEX,
                0.5,
                (255, 0, 255),
                1,
            )
        for x in range(0, width, 50):
            cv2.line(canvas, (x, 0), (x, height), (255, 0, 255), 1)
            cv2.putText(
                canvas,
                str(x),
                (x + 5, 15),
                cv2.FONT_HERSHEY_SIMPLEX,
                0.5,
                (255, 0, 255),
                1,
            )
        cv2.namedWindow("The first frame in coordinates", cv2.WINDOW_NORMAL)
        cv2.imshow("The first frame in coordinates", canvas)

        while True:
            dialog = wx.TextEntryDialog(
                self,
                "Enter the coordinates (integers) of the cropping window",
                "Format:[left,right,top,bottom]",
            )
            if dialog.ShowModal() != wx.ID_OK:
                self.crop_frame = False
                self.text_cropframe.SetLabel("Not to crop the frames")
                dialog.Destroy()
                break

            coordinates = dialog.GetValue().split(",")
            if len(coordinates) != 4:
                self.crop_frame = False
                wx.MessageBox(
                    "Please enter the coordinates (integers) in correct format.",
                    "Error",
                    wx.OK | wx.ICON_ERROR,
                )
                self.text_cropframe.SetLabel("Not to crop the frames")
                continue
            try:
                self.left = int(coordinates[0])
                self.right = int(coordinates[1])
                self.top = int(coordinates[2])
                self.bottom = int(coordinates[3])
            except ValueError:
                self.crop_frame = False
                wx.MessageBox(
                    "Please enter 4 integers.",
                    "Error",
                    wx.OK | wx.ICON_ERROR,
                )
                self.text_cropframe.SetLabel("Not to crop the frames")
                continue

            self.crop_frame = True
            self.text_cropframe.SetLabel(
                f"The cropping window is from left: {self.left} to right: {self.right}, from top: {self.top} to bottom: {self.bottom}."
            )
            break

        cv2.destroyAllWindows()

    def enhance_contrasts(self, event):
        """Opens dialogs to configure frame cropping."""

        if self.video_paths is None:
            wx.MessageBox("No video selected.", "Error", wx.OK | wx.ICON_ERROR)
            return

        capture = cv2.VideoCapture(self.video_paths[0])
        _, frame = capture.read()
        capture.release()

        if self.framewidth is not None:
            frame = cv2.resize(
                frame,
                (
                    self.framewidth,
                    int(frame.shape[0] * self.framewidth / frame.shape[1]),
                ),
                interpolation=cv2.INTER_AREA,
            )

        while True:
            cv2.destroyAllWindows()
            cv2.namedWindow("The first frame in coordinates", cv2.WINDOW_NORMAL)
            cv2.imshow("The first frame in coordinates", frame)

            dialog = wx.TextEntryDialog(
                self,
                "Enter the fold changes for contrast enhancement",
                "A number between 1.0~5.0",
            )
            if dialog.ShowModal() != wx.ID_OK:
                self.enhance_contrast = False
                self.text_enhancecontrast.SetLabel("Not to enhance contrast.")
                dialog.Destroy()
                break

            contrast = dialog.GetValue()
            try:
                self.contrast = float(contrast)
            except ValueError:
                self.enhance_contrast = False
                wx.MessageBox(
                    "Please enter a float number between 1.0~5.0.",
                    "Error",
                    wx.OK | wx.ICON_ERROR,
                )
                self.text_enhancecontrast.SetLabel("Not to enhance contrast.")
                dialog.Destroy()
                continue

            show_frame = frame * self.contrast  # type: ignore
            show_frame[show_frame > 255] = 255
            show_frame = np.uint8(show_frame)
            cv2.destroyAllWindows()
            cv2.namedWindow("The first frame in coordinates", cv2.WINDOW_NORMAL)
            cv2.imshow("The first frame in coordinates", show_frame)  # type: ignore

            confirm = wx.MessageDialog(
                self,
                "Apply the current contrast value?",
                "Apply value?",
                wx.YES_NO | wx.ICON_QUESTION,
            )
            if confirm.ShowModal() != wx.ID_YES:
                self.enhance_contrast = False
                self.text_enhancecontrast.SetLabel("Not to enhance contrast.")
                confirm.Destroy()
                dialog.Destroy()
                continue

            self.enhance_contrast = True
            self.text_enhancecontrast.SetLabel(
                f"The contrast enhancement fold change is: {self.contrast}."
            )
            confirm.Destroy()
            dialog.Destroy()
            cv2.destroyAllWindows()
            break

    @property
    def fps_reduction_factor(self):
        """
        The factor by which to reduce the video framerate.

        The new FPS of the video will be calculated using the formula
        new_fps = old_fps / factor. The reduction factor must be greater than
        1.0.
        """
        return self._fps_reduction_factor

    @fps_reduction_factor.setter
    def fps_reduction_factor(self, factor: float):
        """Ensure reduction factor is greater than or equal to 1.0 before setting."""
        if factor < 1.0:
            raise ValueError("FPS reduction factor must be greater than or equal to 1.")
        self._fps_reduction_factor = factor

    def reduce_fps(self, event):
        """Let user enter FPS reduction factor."""
        while True:
            fps_dialog = wx.TextEntryDialog(
                self,
                "Enter a number greater than 1.0",
                "Enter FPS reduction factor",
            )
            if fps_dialog.ShowModal() != wx.ID_OK:
                break

            try:
                factor = float(fps_dialog.GetValue())
            except ValueError:
                wx.MessageBox(
                    "Please enter a number.",
                    "Error",
                    wx.OK | wx.ICON_ERROR,
                )
                continue
            try:
                self.fps_reduction_factor = factor
            except ValueError as e:
                wx.MessageBox(
                    str(e),
                    "Error",
                    wx.OK | wx.ICON_ERROR,
                )
                continue
            self.text_fps.SetLabel(f"FPS Reduction Factor: {factor}")
            break
        fps_dialog.Destroy()

    def parse_time_window_filename(self, video_path: str) -> list[Tuple[float, float]]:
        """
        Return the time windows to trim the video given the video file.

        The video filename should contain tags st<start> and ed<end> where
        start and end are the start and end times of the time window, and each
        tag is separated from the video file by an underscore.

        Ex. /path/to/video_st0_ed30_st120_ed150.mov will return the time
        windows (0, 30) and (120, 150).

        Args:
            video_path: The path to the video to extract time windows from.

        Returns:
            A list of tuples (start, end) where start and end are floats that
            represent the start and end times in seconds of each time window.

        Raises:
            ValueError: There was an error parsing the file path.
        """
        video = Path(video_path)
        parts = [
            part
            for part in video.stem.split("_")
            if part.startswith("st") or part.startswith("ed")
        ]

        windows = []
        for i in range(0, len(parts), 2):
            if i >= len(parts):
                raise ValueError(f"Missing ed tag in file {video.name}.")
            try:
                assert parts[i].startswith("st")
                start = float(parts[i][2:])
            except (ValueError, AssertionError):
                raise ValueError(f"Invalid tag {parts[i]} in file {video.name}.")
            try:
                assert parts[i + 1].startswith("ed")
                end = float(parts[i + 1][2:])
            except (ValueError, AssertionError):
                raise ValueError(f"Invalid tag {parts[i+1]} in file {video.name}.")
            windows.append((start, end))

        return windows

    def preprocess_videos(self, event):
        """Process videos given the current configuration."""

        if self.video_paths is None or self.result_path is None:
            wx.MessageBox(
                "No input video(s) / output folder.", "Error", wx.OK | wx.ICON_ERROR
            )
            return

        print("Start to preprocess video(s)...")
        for video_path in self.video_paths:
            if self.decode_t is True:
                self.time_windows = self.parse_time_window_filename(video_path)

            preprocess_video(
                video_path,
                self.result_path,
                self.framewidth,
                trim_video=self.trim_video,
                time_windows=self.time_windows,
                enhance_contrast=self.enhance_contrast,
                contrast=self.contrast,
                crop_frame=self.crop_frame,
                left=self.left,
                right=self.right,
                top=self.top,
                bottom=self.bottom,
                fps_reduction_factor=self.fps_reduction_factor,
            )
        print("Preprocessing completed!")
